from fugue_sql_antlr._parser.sa_fugue_sql import USE_CPP_IMPLEMENTATION

FUGUE_SQL_CPP_PARSER_AVAILABLE = USE_CPP_IMPLEMENTATION
FUGUE_CONF_SQL_IGNORE_CASE = "fugue.sql.compile.ignore_case"
FUGUE_CONF_SQL_PARSER = "fugue.sql.compile.parser"
_FUGUE_SQL_CASE_ISSUE_THREHOLD = 0.8



from typing import Iterable, List, Optional, Tuple

from antlr4 import InputStream, Token
from antlr4.Token import CommonToken
from antlr4.tree.Tree import ParseTree, TerminalNode

from fugue_sql_antlr._parser.fugue_sqlParser import fugue_sqlParser
from fugue_sql_antlr._parser.sa_fugue_sql import (
    SA_ErrorListener,
    _cpp_parse,
    _py_parse,
    parse,
)
from fugue_sql_antlr.constants import (
    _FUGUE_SQL_CASE_ISSUE_THREHOLD,
    FUGUE_CONF_SQL_IGNORE_CASE,
    FUGUE_SQL_CPP_PARSER_AVAILABLE,
)


class FugueSQLParser:
    def __init__(
        self,
        code: str,
        rule: str,
        ignore_case: bool,
        parse_mode: str,
    ):
        self._rule = rule
        self._raw_code = code
        self._raw_lines = code.splitlines()
        self._parse_mode = parse_mode
        self._ignore_case = ignore_case
        if ignore_case:
            self._tree = self._to_tree(upper=True)
        else:
            try:
                self._tree = self._to_tree(upper=False)
            except SyntaxError as e:
                if _detect_case_issue(code, _FUGUE_SQL_CASE_ISSUE_THREHOLD):
                    prefix = (
                        "(FugueSQL requires uppercase characters by default. "
                        f"To ignore casing, turn on {FUGUE_CONF_SQL_IGNORE_CASE})"
                    )
                    msg = prefix + str(e)
                    raise SyntaxError(msg).with_traceback(None) from None
                else:
                    raise SyntaxError(str(e)) from None
        self._tokens = list(self._get_tokens(self.tree))
        for i, token in enumerate(self._tokens):
            token._arr_pos = i

    @property
    def code(self) -> str:
        return self._raw_code

    @property
    def tree(self) -> ParseTree:
        return self._tree

    def get_norm_text(
        self,
        node: Optional[ParseTree],
        delimiter: str = " ",
        upper_keyword: bool = False,
    ) -> str:
        def _get_token_str(token: Token):
            s = self.code[token.start : token.stop + 1]
            if upper_keyword and self._ignore_case and hasattr(token, "is_keyword"):
                return s.upper()
            return s

        if node is None:
            return ""
        tp = self._get_range(node)
        if tp is None:
            return ""
        if tp[0] is tp[1]:
            return self.code[tp[0].start : tp[0].stop + 1]
        return delimiter.join(
            _get_token_str(self._tokens[i])
            for i in range(tp[0]._arr_pos, tp[1]._arr_pos + 1)
        )

    def get_raw_text(self, node: Optional[ParseTree], add_lineno: bool = False) -> str:
        if node is None:
            return ""
        tp = self._get_range(node)
        if tp is None:
            return ""
        if not add_lineno:
            return self._raw_code[tp[0].start : tp[1].stop + 1]
        start = tp[0].line
        stop = tp[1].line
        lines: List[str] = []
        while start <= stop:
            prefix = "" if not add_lineno else f"{start}:\t"
            lines.append(prefix + self._raw_lines[start - 1])
            start += 1
        return "\n".join(lines)

    def _get_tokens(
        self, node: Optional[ParseTree], parent: Optional[ParseTree] = None
    ) -> Iterable[CommonToken]:
        if node is None:
            return
        elif isinstance(node, Token):
            if isinstance(node, CommonToken) and node.type != Token.EOF:
                if not isinstance(
                    parent, fugue_sqlParser.NonReservedContext
                ) and hasattr(parent, node.text):
                    node.is_keyword = True
                yield node
        elif isinstance(node, TerminalNode):
            yield from self._get_tokens(node.getSymbol(), parent)
        else:
            for i in range(node.getChildCount()):
                for x in self._get_tokens(node.getChild(i), node):
                    yield x

    def _get_range(self, node: ParseTree) -> Optional[Tuple[CommonToken, CommonToken]]:
        if isinstance(node, CommonToken):
            return node, node
        s = node.start
        if not isinstance(s, CommonToken) or s.type == Token.EOF:
            return None
        return s, node.stop

    def _to_tree(self, upper: bool) -> ParseTree:
        stream = InputStream(self._raw_code if not upper else self._raw_code.upper())
        handler = _ErrorListener(self)
        if self._parse_mode == "auto":
            return parse(stream, self._rule, sa_err_listener=handler)
        if self._parse_mode == "cpp":
            if not FUGUE_SQL_CPP_PARSER_AVAILABLE:
                raise ValueError(
                    "CPP parser is not installed, use 'auto' or 'py' as parse_mode"
                )
            return _cpp_parse(stream, self._rule, sa_err_listener=handler)
        if self._parse_mode == "py":
            return _py_parse(stream, self._rule, sa_err_listener=handler)
        raise ValueError(f"{self._parse_mode} is not valid")


def _detect_case_issue(text: str, lower_case_percentage: float) -> bool:
    letters, lower = 0, 0.0
    for c in text:
        if "a" <= c <= "z":
            lower += 1.0
            letters += 1
        elif "A" <= c <= "Z":
            letters += 1
    if letters == 0:
        return False
    return lower / letters >= lower_case_percentage


class _ErrorListener(SA_ErrorListener):
    def __init__(self, parser: FugueSQLParser):
        super().__init__()
        self._parser = parser

    def syntaxError(
        self,
        input_stream: InputStream,
        offendingSymbol: Token,
        char_index: int,
        line: int,
        column: int,
        msg: str,
    ):
        message = self._parser.get_raw_text(offendingSymbol, add_lineno=True)
        s = message.split("\t", 1)
        pre = "".join([" "] * len(s[0])) + "\t"
        dots = "".join(["."] * column)
        arrows = "".join(["^"] * (offendingSymbol.stop - offendingSymbol.start + 1))
        raise SyntaxError(f"\n{message}\n{pre}{dots}{arrows}")



# This file was auto-generated by speedy-antlr-tool v1.4.1
#  https://github.com/amykyta3/speedy-antlr-tool

import sys
import types

import antlr4
from antlr4 import InputStream, CommonTokenStream, Token
from antlr4.tree.Tree import ParseTree
from antlr4.error.ErrorListener import ErrorListener

from .fugue_sqlParser import fugue_sqlParser
from .fugue_sqlLexer import fugue_sqlLexer

#-------------------------------------------------------------------------------
# User API
#-------------------------------------------------------------------------------
#: Defines whether C++ implementation is used when calling parse()
#: This is automatically set to False if the accelerator is not available.
#: You may override this to False to force use of Python fallback implementation.
USE_CPP_IMPLEMENTATION = True


class SA_ErrorListener:
    """
    Base callback class to handle Antlr syntax errors.

    Not able to do a 1-to-1 bridge of Antlr's error listener
    Instead, this class provides roughly equivalent functionality.
    """

    def syntaxError(self, input_stream:InputStream, offendingSymbol:Token, char_index:int, line:int, column:int, msg:str):
        """
        Called when lexer or parser encountered a syntax error.

        Parameters
        ----------
        input_stream:InputStream
            Reference to the original input stream that this error is from

        offendingSymbol:Token
            If available, denotes the erronous token

        char_index:int
            Character offset of the error within the input stream

        line:int
            Line number of the error

        column:int
            Character offset within the line

        msg:str
            Antlr error message
        """
        pass


def parse(stream:InputStream, entry_rule_name:str, sa_err_listener:SA_ErrorListener=None) -> ParseTree:
    """
    Parse the input stream

    If C++ implementation of parser is not available, automatically falls back
    to Python implementation.

    Parameters
    ----------
    stream:InputStream
        Source stream to lex & parse

    entry_rule_name:str
        Name of grammar rule to use as the entry point

    sa_err_listener:SA_ErrorListener
        Optionally override the error listener.
        By default, Antlr's default ConsoleErrorListener is used

        Important! This is NOT a reference to an Antlr ErrorListener class!
        This is a slightly different implementation.
    """
    if USE_CPP_IMPLEMENTATION:
        return _cpp_parse(stream, entry_rule_name, sa_err_listener)
    else:
        return _py_parse(stream, entry_rule_name, sa_err_listener)


#-------------------------------------------------------------------------------
# C++ implementation of parser
#-------------------------------------------------------------------------------

try:
    from fugue_sql_antlr_cpp import sa_fugue_sql_cpp_parser
except ImportError:
    USE_CPP_IMPLEMENTATION = False

def _cpp_parse(stream:InputStream, entry_rule_name:str, sa_err_listener:SA_ErrorListener=None) -> ParseTree:
    # Validate input types here before handing over to C++
    if not isinstance(stream, InputStream):
        raise TypeError("'stream' shall be an Antlr InputStream")
    if not isinstance(entry_rule_name, str):
        raise TypeError("'entry_rule_name' shall be a string")
    if sa_err_listener is not None and not isinstance(sa_err_listener, SA_ErrorListener):
        raise TypeError("'sa_err_listener' shall be an instance of SA_ErrorListener or None")

    return sa_fugue_sql_cpp_parser.do_parse(fugue_sqlParser, stream, entry_rule_name, sa_err_listener)


#-------------------------------------------------------------------------------
# Fall-back Python implementation of parser
#-------------------------------------------------------------------------------

class _FallbackErrorTranslator(ErrorListener):
    """
    Translates syntax error to user-defined SA_ErrorListener callback
    """
    def __init__(self, sa_err_listener:SA_ErrorListener, input_stream:InputStream):
        self.sa_err_listener = sa_err_listener
        self.input_stream = input_stream

    def syntaxError(self, recognizer, offendingSymbol, line, column, msg, e):
        if isinstance(recognizer, antlr4.Lexer):
            input_stream = recognizer.inputStream
        elif isinstance(recognizer, antlr4.Parser):
            input_stream = recognizer.getInputStream()
        else:
            raise RuntimeError("Unknown recognizer type")

        char_index = input_stream.index

        self.sa_err_listener.syntaxError(
            self.input_stream, offendingSymbol, char_index, line, column, msg
        )


def _py_parse(stream:InputStream, entry_rule_name:str, sa_err_listener:SA_ErrorListener=None) -> ParseTree:
    if sa_err_listener is not None:
        err_listener = _FallbackErrorTranslator(sa_err_listener, stream)

    # Lex
    lexer = fugue_sqlLexer(stream)
    if sa_err_listener is not None:
        lexer.removeErrorListeners()
        lexer.addErrorListener(err_listener)
    token_stream = CommonTokenStream(lexer)

    # Parse
    parser = fugue_sqlParser(token_stream)
    if sa_err_listener is not None:
        parser.removeErrorListeners()
        parser.addErrorListener(err_listener)

    entry_rule_func = getattr(parser, entry_rule_name, None)
    if not isinstance(entry_rule_func, types.MethodType):
        raise ValueError("Invalid entry_rule_name '%s'" % entry_rule_name)
    return entry_rule_func()


# flake8: noqa
from fugue_sql_antlr.constants import (
    FUGUE_CONF_SQL_IGNORE_CASE,
    FUGUE_CONF_SQL_PARSER,
    FUGUE_SQL_CPP_PARSER_AVAILABLE,
)
from fugue_sql_antlr.parser import FugueSQLParser
from fugue_sql_antlr.visitor import FugueSQLVisitor
